# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flexible_list_of_values',
 'flexible_list_of_values.management',
 'flexible_list_of_values.management.commands',
 'flexible_list_of_values.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.2']

setup_kwargs = {
    'name': 'flexible-list-of-values',
    'version': '0.2.0',
    'description': 'Flexible Lists of Values (LOV) for Django',
    'long_description': '# flexible-list-of-values\n\nFlexible and Extensible Lists of Values (LOV) for Django\n\n> When adding customizability to a SaaS app, there are a variety of approaches and tools:\n> \n> - [Dynamic models](https://baserow.io/blog/how-baserow-lets-users-generate-django-models) and dynamic model fields\n> - [Entity-Attribute-Value (EAV)](https://en.wikipedia.org/wiki/Entity%E2%80%93attribute%E2%80%93value_model)\n> - Adding [JSONField](https://docs.djangoproject.com/en/4.1/ref/models/fields/#jsonfield) to a model\n>\n> But these approaches can be a bit *too* flexible. Sometimes we want to provide guardrails for our tenants.\n>\n> *flexible-list-of-values* provides you a way to give your SaaS app\'s tenants customization options in its User-facing forms, but also lets you provide them with defaults - either mandatory or optional - to prevent each tenant having to "recreate the wheel".\n\n*Note: The terms "entity" and "tenant" are used interchangeably in this document to refer to a model within your project that has associated Users, and which has "ownership" over the LOV value options provided to its Users*\n\n## Example implementation\n\nImagine your project provides a form that allows your tenants\' users provide details about their garden plots. Now you want to add a "crops" field to this form so that each tenants\' users can also record the specific crops they are growing.\n\nYou could hard-code some crop value choices using CharField and choices, or if your tenants wanted the field to be customizable you could use JSONField or other methods, but this can get sloppy.\n\n`flexible-list-of-values` provides an in-between option. You can specify some default options that are mandatory for all users, regardless of tenant, provide some default options that your tenants can either use or discard, and allow additional custom value options specific to the tenant and its users.\n\n![Descriptive Diagram](https://raw.githubusercontent.com/jacklinke/flexible-list-of-values/main/docs/media/FlexibleLOVDescription.png)\n\n\n### Docker Compose\n\nThis package comes with a demonstration project using docker compose. In order to utilize the project, we assume here that you have downloaded or cloned a copy of the [project code repository](https://github.com/jacklinke/flexible-list-of-values), and that you already have docker compose installed. If not, see [the compose docs](https://docs.docker.com/compose/install/).\n\nNavigate to the "flexible-list-of-values" directory, and create a virtual environment to work from:\n\n`python3 -m venv .venv`\n    \nThen activate the environment:\n\n`source .venv/bin/activate`\n\nBuild the project:\n\n`docker compose build`\n\nThen run migrations and create a superuser.\n\n```bash\npython manage.py migrate\npython manage.py createsuperuser\n```\n\nTo synchronize the "lov_default" values with the database, run\n\n`python manage.py update_lovs`\n\nBring the docker compose project up, where it will be available at http://127.0.0.1:8000/\n\n`docker compose up -d`\n\nOptionally create another user or two in [admin](http://127.0.0.1:8000/admin/)\n\nWithin admin, create at least one tenant, setting the owner and users fields. The tenant owner can create new LOV Values and choose which Values its Users can select from.\n\nIf the user account you are logged in with is a tenant owner, then [click here](http://127.0.0.1:8000/lov_crop_value_create_view/) to add new LOV Values, or [click here](http://127.0.0.1:8000/lov_tenant_crop_selection_view/) to specify which Values your users can choose from. \n\nIf the user account you are logged in with belongs to a tenant, then [click here](http://127.0.0.1:8000/lov_user_crop_selection_view/) to choose from among the LOV Value choices the tenant has made available to you.\n\n*See the Test Project in the [project repo](https://github.com/jacklinke/flexible-list-of-values) for full details.*\n\n### The Testapp Project\n\nHere is what is going on behind the scenes.\n\nmodels.py:\n\n```python\nfrom django.contrib.auth import get_user_model\nfrom django.db import models\nfrom django.utils.translation import gettext_lazy as _\nfrom flexible_list_of_values import LOVValueType\nfrom flexible_list_of_values.models import AbstractLOVSelection, AbstractLOVValue\n\nUser = get_user_model()\n\n\nclass Tenant(models.Model):\n    """\n    This model is a very simplistic example of how one might implement a Tenant architecture.\n    """\n    name = models.CharField(_("Tenant Name"), max_length=100)\n    owner = models.ForeignKey(User, on_delete=models.CASCADE, related_name="owned_tenants", null=True, blank=True)\n\n    users = models.ManyToManyField(User, related_name="tenants", blank=True)\n\n    def __str__(self):\n        return self.name\n\n\nclass TenantCropLOVValue(AbstractLOVValue):\n    """\n    Concrete implementation of AbstractLOVValue with Crop options that a Tenant can modify.\n\n    "Fruit", "Herbs and Spices", and "Vegetable" are mandatory selections, but the others are provided to the Tenants as optional recommendations. Tenants can also add their own custom Values.\n    """\n    lov_entity_model = "testapp.Tenant"\n    lov_selections_model = "testapp.TenantCropLOVSelection"\n\n    lov_defaults = {\n        "Fruit": {"value_type": LOVValueType.MANDATORY},  # 1\n        "Fruit - Apple": {"value_type": LOVValueType.OPTIONAL},\n        "Fruit - Mellon": {"value_type": LOVValueType.OPTIONAL},\n        "Fruit - Stone": {"value_type": LOVValueType.OPTIONAL},\n        "Herbs and Spices": {"value_type": LOVValueType.MANDATORY},  # 2\n        "Herbs and Spices - Basil": {"value_type": LOVValueType.OPTIONAL},\n        "Herbs and Spices - Sage": {"value_type": LOVValueType.OPTIONAL},\n        "Herbs and Spices - Thyme": {"value_type": LOVValueType.OPTIONAL},\n        "Vegetable": {"value_type": LOVValueType.MANDATORY},  # 3\n        "Vegetable - Avocado": {"value_type": LOVValueType.OPTIONAL},\n        "Vegetable - Cabbage": {"value_type": LOVValueType.OPTIONAL},\n        "Vegetable - Spinach": {"value_type": LOVValueType.OPTIONAL},\n    }\n\n    class Meta(AbstractLOVValue.Meta):\n        verbose_name = "Tenant Crop Value"\n        verbose_name_plural = "Tenant Crop Values"\n        ordering = ["name"]\n\n\nclass TenantCropLOVSelection(AbstractLOVSelection):\n    """\n    Concrete implementation of AbstractLOVSelection with actual selections\n        of Crops that a Tenant\'s Users can select from in Forms\n    """\n    lov_value_model = "testapp.TenantCropLOVValue"\n    lov_entity_model = "testapp.Tenant"\n\n    class Meta(AbstractLOVSelection.Meta):\n        verbose_name = "Tenant Crop Selection"\n        verbose_name_plural = "Tenant Crop Selections"\n\n    def __str__(self):\n        return self.lov_value.name\n\n\nclass UserCrop(models.Model):\n    """\n    The crops which a User belonging to a particular Tenant has selected for their garden plot. (This is a very simplistic implementation of a Tenant-specific model and does not represent best practices.)\n    """\n\n    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="user_crops")\n    tenant = models.ForeignKey(Tenant, on_delete=models.CASCADE, related_name="user_crops")\n\n    crops = models.ManyToManyField(TenantCropLOVValue)\n```\n\nFor the example above, in a UserCrop form for Users of TenantA or TenantB ...\n\n- The `crops` ModelChoiceField will always include Crop choices for "Fruit", "Herbs and Spices", and "Vegetable".\n- Both tenants can choose whether their Users will have the other values listed in `lov_defaults`.\n- And both tenants can create custom tenant-specific value options for their Users to choose from.\n\n## Working with LOV Values\n\nView the available Values for this Tenant:\n\n```python\ntenant = Tenant.objects.first()\nvalues_for_tenant = TenantCropLOVValue.objects.for_entity(tenant)\n```\n\nAlternately:\n\n```python\ntenant = Tenant.objects.first()\nvalues_for_tenant = TenantCropLOVSelection.objects.values_for_entity(tenant)\n```\n\nView the selected Values for this Tenant:\n\n```python\ntenant = Tenant.objects.first()\nselected_values_for_entity = TenantCropLOVSelection.objects.selected_values_for_entity(tenant)\n```\n\nCreate new custom Values for this Tenant:\n\n```python\ntenant = Tenant.objects.first()\nTenantCropLOVValue.objects.create_for_entity(tenant, "New Value for this Tenant")\n```\n\nDelete Values for this Tenant (only deletes custom values owned by this tenant)\n\n```python\ntenant = Tenant.objects.first()\nvalues = TenantCropLOVValue.objects.for_entity(entity=tenant).filter(name__icontains="Something")\nfor value in values:\n    value.delete()\n```\n\n\n## Letting tenants select LOV value choices for their users\n\nTenants can select from among the Values available for this Tenant or create new Values\n\nforms.py:\n\n```python\nfrom django import forms\nfrom flexible_list_of_values.forms import (\n    LOVValueCreateFormMixin,\n    LOVSelectionsModelForm,\n)\nfrom tests.testapp.models import TenantCropLOVSelection, TenantCropLOVValue\n\n\nclass TenantCropValueCreateForm(LOVValueCreateFormMixin, forms.ModelForm):\n    """\n    Form to let a tenant add a new custom LOV Value.\n    """\n\n    class Meta:\n        model = TenantCropLOVValue\n        fields = ["name", "lov_entity", "value_type"]\n\n\nclass TenantCropValueSelectionForm(LOVSelectionsModelForm):\n    """\n    Form to let a tenant select which LOV Values its users can choose from.\n    """\n\n    class Meta:\n        model = TenantCropLOVSelection\n\n    def __init__(self, *args, **kwargs):\n        super().__init__(*args, **kwargs)\n        self.fields["lov_selections"].widget.attrs["size"] = "10"\n```\n\nviews.py\n\n```python\nfrom django.template.response import TemplateResponse\nfrom tests.testapp.models import TenantCropLOVValue\nfrom tests.testapp.forms import TenantCropValueCreateForm, TenantCropValueSelectionForm\n\n\n# Add decorator or other logic to allow only logged in tenant owners access to this view\ndef lov_crop_value_create_view(request):\n    """\n    Form for creating new values for a Tenant\n    """\n\n    template = "testapp/create_value.html"\n    context = {}\n\n    # However you specify the current entity/tenant for the User submitting this form.\n    # This is only an example.\n    tenant = request.user.owned_tenants.first()\n\n    # Here we provide the User\'s entity, which the form will use to determine the available Values\n    form = TenantCropValueCreateForm(request.POST or None, lov_entity=tenant)\n\n    if request.method == "POST":\n        if form.is_valid():\n            form.save()\n\n    context["form"] = form\n\n    # Provide the list of existing LOV Values for this Tenant\n    context["existing_values"] = TenantCropLOVValue.objects.for_entity(tenant)\n\n    return TemplateResponse(request, template, context)\n\n\n# Add decorator or other logic to allow only logged in tenant owners access to this view\ndef lov_tenant_crop_selection_view(request):\n    """\n    Form for selecting the Values a tenant wants to use\n    """\n\n    template = "testapp/select_values.html"\n    context = {}\n\n    # However you specify the current entity/tenant for the User submitting this form.\n    # This is only an example.\n    tenant = request.user.owned_tenants.first()\n\n    # Here we provide the entity\n    form = TenantCropValueSelectionForm(request.POST or None, lov_entity=tenant)\n\n    if request.method == "POST":\n        if form.is_valid():\n            form.save()\n            # Update form\'s contents to ensure mandatory items are selected\n            form = TenantCropValueSelectionForm(None, lov_entity=tenant)\n\n    context["form"] = form\n\n    return TemplateResponse(request, template, context)\n```\n\n## Working with a tenant\'s LOV Selections\n\nA Tenant\'s Users make a choice from among the selected Values for this Tenant each time they fill out a UserCrop Selection Form.\n\nforms.py (continued from previous forms.py code):\n\n```python\nfrom django import forms\nfrom tests.testapp.models import UserCrop\n\nclass UserCropSelectionForm(forms.ModelForm):\n    """\n    Form to let a Tenant\'s Users select from available LOV Values.\n    """\n\n    class Meta:\n        model = UserCrop\n        fields = "__all__"\n\n    def __init__(self, *args, **kwargs):\n        self.user = kwargs.pop("user", None)\n        super().__init__(*args, **kwargs)\n        # Get all allowed values for this tenant\n        if self.instance:\n            self.fields["crops"].queryset = TenantCropLOVSelection.objects.selections_for_entity(self.instance.tenant)\n            self.fields["user"].initial = self.user\n            self.fields["tenant"].initial = self.instance.tenant\n        else:\n            self.fields["crops"].queryset = TenantCropLOVSelection.objects.none()\n\n        self.fields["crops"].widget.attrs["size"] = "10"\n```\n\nviews.py (continued from previous views.py code):\n\n```python\nfrom django.template.response import TemplateResponse\nfrom tests.testapp.models import UserCrop\nfrom tests.testapp.forms import UserCropSelectionForm\n\n# Add decorator or other logic to allow only logged in users who belong to a tenant to access to this view\ndef lov_user_crop_selection_view(request):\n    """\n    A for that allows a Tenant\'s Users to select the Crops they want to select\n    """\n\n    template = "testapp/select_values.html"\n    context = {}\n\n    # However you specify the current entity/tenant for the User submitting this form.\n    # This is only an example.\n    tenant = request.user.tenants.first()\n\n    obj = obj, created = UserCrop.objects.get_or_create(\n        user=request.user,\n        tenant=tenant,\n    )\n\n    # Here we provide the entity\n    form = UserCropSelectionForm(request.POST or None, instance=obj)\n    if request.method == "POST":\n        if form.is_valid():\n            form.save()\n\n    context["form"] = form\n    return TemplateResponse(request, template, context)\n```\n\nHere, Users for TenantA who are filling out a UserCropSelectionForm form will see all mandatory values, the optional values that TenantA has selected, and any custom values TenantA has created. TenantB\'s users will see all mandatory values, the optional values that TenantB has selected, and any custom values TenantB has created.\n\n## Management Commands\n\n- `update_lovs`: Synchronizes the `lov_defaults` in each model, if any, with the database.\n\n## API\n\n### Model: AbstractLOVValue\n\n#### Fields\n\n- **id**: default id\n- **name** (CharField): The name or title of the value to be used.\n- **lov_entity** (FK): the owning entity for this value. If this is a default value, this field will be null.\n- **lov_associated_entities** (M2M): all entities this value is associated with. (The reverse relationship on the entity model is all values selected for the entity)\n- **value_type** (CharField): Any of\n    - LOVValueType.MANDATORY\n    - LOVValueType.OPTIONAL\n    - LOVValueType.CUSTOM\n- **deleted** (DateTimeField): The datetime this value was deleted, or null if it is not deleted.\n\n\n#### Model attributes\n\n<dl>\n  <dt>lov_defaults</dt>\n  <dd>\n    A dictionary of default mandatory and optional values from which an entity can select. See usage examples above.<br>\n    <em>Default</em>: <code>{}</code>\n  </dd>\n\n  <dt>lov_entity_model</dt>\n  <dd>\n    Specifies the model class for the \'entity\' in your project which can customize its Users\' LOVs. Specify the string representation of the model class (e.g.: <code>"entities.Entity"</code>).<br>\n    <em>* Required</em>\n  </dd>\n\n  <dt>lov_entity_on_delete</dt>\n  <dd>\n    What should happen when the related entity instance is deleted.<br>\n    <em>Default</em>: <code>models.CASCADE</code>\n  </dd>\n\n  <dt>lov_entity_model_related_name</dt>\n  <dd>\n    <code>related_name</code> for the related entity instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)s_related"</code>\n  </dd>\n\n  <dt>lov_entity_model_related_query_name</dt>\n  <dd>\n    <code>related_query_name</code> for the related entity instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)ss"</code>\n  </dd>\n\n  <dt>lov_selections_model</dt>\n  <dd>\n    Specifies the model class of the through-model between an Entity and a Value. Each instance of this through-model is an option that the tenant\'s users can choose from. Must be a concrete subclass of AbstractLOVSelection. Specify the string representation of the model class (e.g.: <code>"entities.TenantCropLOVSelection"</code>).<br>\n    <em>* Required</em>\n  </dd>\n\n  <dt>lov_associated_entities_related_name</dt>\n  <dd>\n    <code>related_name</code> for the M2M to the entity instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)s_related"</code>\n  </dd>\n\n  <dt>lov_associated_entities_related_query_name</dt>\n  <dd>\n    <code>related_query_name</code> for the M2M to the entity instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)ss"</code>\n  </dd>\n</dl>\n\n#### Manager and QuerySet Methods\n\n<dl>\n  <dt>for_entity(entity)</dt>\n  <dd>\n    Returns QuerySet of all available values for a given entity, including:<br>\n    <ul>\n        <li>all required default values</li>\n        <li>all non-required default values</li>\n        <li>all entity-specific values for this entity</li>\n    </ul>\n  </dd>\n\n  <dt>create_for_entity(entity, name: str)</dt>\n  <dd>\n    Creates a new selectable Value for the provided entity.\n  </dd>\n\n  <dt>create_mandatory(name: str)</dt>\n  <dd>\n    Creates a new Value (selected for all entities).\n  </dd>\n\n  <dt>create_optional(name: str)</dt>\n  <dd>\n    Creates a new selectable optional Value (selectable by all entities).\n  </dd>\n</dl>\n\n\n### Model: AbstractLOVSelection\n\nThis is a through-model from an concrete LOVValue model instance to an entity model instance representing the value selections an entity has made.\n\n#### Fields\n\n- **id**: default id\n- **lov_entity** (FK): the entity this selection is associated with.\n- **lov_value** (FK): the value this selection is associated with.\n\n#### Model attributes\n\n<dl>\n  <dt>lov_entity_model</dt>\n  <dd>\n    Specifies the model class for the \'entity\' in your project which can customize its Users\' LOVs. Specify the string representation of the model class (e.g.: <code>"entities.Entity"</code>).<br>\n    <em>* Required</em>\n  </dd>\n\n  <dt>lov_entity_on_delete</dt>\n  <dd>\n    What should happen when the related entity instance is deleted.<br>\n    <em>Default</em>: <code>models.CASCADE</code>\n  </dd>\n\n  <dt>lov_entity_model_related_name</dt>\n  <dd>\n    <code>related_name</code> for the related entity instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)s_related"</code>\n  </dd>\n\n  <dt>lov_entity_model_related_query_name</dt>\n  <dd>\n    <code>related_query_name</code> for the related entity instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)ss"</code>\n  </dd>\n\n\n  <dt>lov_value_model</dt>\n  <dd>\n    Specifies the model class for the concrete (not abstract!) subclass of AbstractLOVValue. Specify the string representation of the model class (e.g.: <code>"contacts.TenantCropLOVSelection"</code>).<br>\n    <em>* Required</em>\n  </dd>\n\n  <dt>lov_value_model_related_name</dt>\n  <dd>\n    <code>related_name</code> for the related concrete subclassed AbstractLOVValue instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)s_related"</code>\n  </dd>\n\n  <dt>lov_value_model_related_query_name</dt>\n  <dd>\n    <code>related_query_name</code> for the related concrete subclassed AbstractLOVValue instance.<br>\n    <em>Default</em>: <code>"%(app_label)s_%(class)ss"</code>\n  </dd>\n</dl>\n\n#### Manager and QuerySet Methods\n\n<dl>\n  <dt>values_for_entity(entity)</dt>\n  <dd>\n    Returns QuerySet of all <em>available</em> values for a given entity, including:<br>\n    <ul>\n        <li>all required default values</li>\n        <li>all non-required default values</li>\n        <li>all entity-specific values for this entity</li>\n    </ul>\n  </dd>\n  \n  <dt>selected_values_for_entity(entity)</dt>\n  <dd>\n    Returns QuerySet of all <em>selected</em> values for a given entity, including:<br>\n    <ul>\n        <li>all required default values</li>\n        <li>all <em>selected</em> non-required default values</li>\n        <li>all <em>selected</em> entity-specific values for this entity</li>\n    </ul>\n  </dd>\n</dl>\n',
    'author': 'Jack Linke',
    'author_email': 'jack@watervize.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jacklinke/flexible-list-of-values',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
